from typing import List

from pyfactcast.client.sync import FactStore


fact_store = FactStore()


def namespaces() -> List[str]:
    with fact_store as fs:
        return fs.enumerate_namespaces()


def types(namespace: str) -> List[str]:
    with fact_store as fs:
        return fs.enumerate_types(namespace)



import typer
from rich.console import Console
import logging
from rich.logging import RichHandler

from .enumerate import app as enum_app
from .streams import app as sub_app
from .fact import app as fac_app


console = Console()
app = typer.Typer()

FORMAT = "%(message)s"


@app.callback()
def main(verbose: bool = False):
    level = "WARNING"
    if verbose:
        level = "DEBUG"

    logging.basicConfig(
        level=level, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
    )


app.add_typer(enum_app, name="enumerate")
app.add_typer(sub_app, name="streams")
app.add_typer(fac_app, name="fact")


typer_click_object = typer.main.get_command(app)

if __name__ == "__main__":
    app()



import logging
import grpc
import struct
from types import TracebackType
from typing import Iterable, List, Optional, Type
from json import dumps
from uuid import UUID

from pyfactcast.grpc.generated.FactStore_pb2_grpc import RemoteFactStoreStub
from pyfactcast.grpc.generated.FactStore_pb2 import (
    MSG_Empty,
    MSG_String,
    MSG_SubscriptionRequest,
    MSG_Notification,
    MSG_UUID,
)

from pyfactcast.client.auth.basic import BasicAuth
from pyfactcast.client.config import (
    ClientConfiguration,
    get_client_configuration,
    log_level,
)
from pyfactcast.client.entities import Fact

log = logging.getLogger()
log.setLevel(log_level)


def get_synchronous_grpc_client(
    client_configuration: Optional[ClientConfiguration] = None,
) -> RemoteFactStoreStub:

    if not client_configuration:
        client_configuration = get_client_configuration()

    options = None
    if client_configuration.ssl_target_override:
        options = (
            (
                "grpc.ssl_target_name_override",
                client_configuration.ssl_target_override,
            ),
        )

    root_cert = None
    if client_configuration.root_cert_path:
        with open(client_configuration.root_cert_path) as f:
            root_cert = f.read().encode("UTF-8")

    channel_credentials = grpc.ssl_channel_credentials(root_certificates=root_cert)

    call_credentials = _construct_call_credentials(client_configuration)

    if call_credentials:
        grpc_credentials = grpc.composite_channel_credentials(
            channel_credentials, call_credentials
        )
    else:
        grpc_credentials = channel_credentials

    channel = grpc.secure_channel(
        target=client_configuration.server,
        credentials=grpc_credentials,
        options=options,
    )

    return RemoteFactStoreStub(channel)


def _construct_call_credentials(
    client_configuration: ClientConfiguration,
) -> Optional[grpc.CallCredentials]:

    if client_configuration.credentials:  # 3.8 Walrus
        call_credentials = grpc.metadata_call_credentials(
            BasicAuth(
                client_configuration.credentials.username,
                client_configuration.credentials.password,
            )
        )
        return call_credentials
    return None


def _fact_filter(msg: MSG_Notification) -> bool:
    if msg.type == msg.Fact:
        return True
    elif msg.type == msg.Catchup:
        log.info("Caught up!")
    elif msg.type == msg.facts:
        log.error("Multiple facts per Notification not implemented.")
    elif msg.type == msg.Complete:
        log.debug("Received stream complete.")
    else:
        log.debug(f"Received Notification of type: {msg.type}")
    return False


class FactStore:
    def __init__(self, client: Optional[RemoteFactStoreStub] = None) -> None:
        self._client = client

    def __enter__(self) -> "FactStore":
        if not self._client:
            self.client = get_synchronous_grpc_client()
        else:
            self.client = self._client
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        # TODO: Implement proper channel termination
        pass

    def enumerate_namespaces(self) -> List[str]:
        res: List[str] = self.client.enumerateNamespaces(MSG_Empty()).embeddedString

        return res

    def enumerate_types(self, namespace: str) -> List[str]:
        message = MSG_String()
        message.embeddedString = namespace
        res: List[str] = self.client.enumerateTypes(message).embeddedString

        return res

    def subscribe(
        self,
        *,
        namespace: str,
        continuous: bool = False,
        from_now: bool = False,
        after_fact: UUID = None,
    ) -> Iterable[Fact]:

        subscription_request_body = {
            "specs": [{"ns": namespace}],
            "continuous": continuous,
        }

        if from_now:
            subscription_request_body["ephemeral"] = True
            subscription_request_body["continuous"] = True

        if after_fact:
            subscription_request_body["startingAfter"] = str(after_fact)

        msg = MSG_SubscriptionRequest(json=dumps(subscription_request_body))
        res = self.client.subscribe(msg)

        return map(Fact.from_msg, filter(_fact_filter, res))

    def serial_of(self, *, fact_id: UUID) -> Optional[str]:
        msb, lsb = struct.unpack(">qq", fact_id.bytes)
        msg = MSG_UUID()
        msg.lsb = lsb
        msg.msb = msb
        res = self.client.serialOf(msg)

        if res.present:
            return str(res.serial)
        return None



from typing import Iterable
from uuid import UUID
import typer
from rich.console import Console
from rich.table import Table
from datetime import datetime
import json

import pyfactcast.app.business.streams as business

app = typer.Typer()
console = Console()


def print_iterable_as_json_list(
    iterable: Iterable,
) -> None:
    console.print("[")
    it = iter(iterable)
    last = next(it)  # type: ignore
    for elem in it:
        console.print(f"{last.json(indent=2)},", soft_wrap=True)
        last = elem
    console.print(last.json(indent=2), soft_wrap=True)
    console.print("]")


@app.command()
def subscribe(
    namespace: str,
    follow: bool = typer.Option(
        False, "--follow", help="Stay connected and listen for new events."
    ),
    from_now: bool = typer.Option(
        False, "--from-now", help="Follow the event stream starting now."
    ),
    after_fact: UUID = typer.Option(
        None, help="Get all events in the stream that happened after a certain ID."
    ),
    json_format: bool = typer.Option(False, "--json", help="Output raw json events."),
) -> None:
    subscription = business.subscribe(
        namespace=namespace, follow=follow, after_fact=after_fact, from_now=from_now
    )
    """
    Subscribe to an eventstream.
    """

    if json_format:
        print_iterable_as_json_list(subscription)
        return

    for elem in subscription:
        table = Table(show_header=False, show_lines=True, header_style="bold magenta")
        table.add_column("Time")
        table.add_column("Type")
        table.add_column("Header")
        table.add_column("Payload")
        table.add_row(
            datetime.fromtimestamp(int(elem.header.meta["_ts"][:10])).isoformat(),  # type: ignore
            elem.header.type,
            elem.header.json(indent=2),
            json.dumps(elem.payload, indent=2),
        )
        console.print(table)



from uuid import UUID
import typer
from rich.console import Console

import pyfactcast.app.business.fact as business

app = typer.Typer()
console = Console()


@app.command()
def serial_of(fact_id: UUID) -> None:
    """
    Returns the serial for the fact identified by the given UUID
    """
    serial = business.serial_of(fact_id)
    if serial:  # 3.8 Walrus
        console.print(serial)
    else:
        console.print(f"No fact with id {fact_id} found.")



from uuid import UUID
from pyfactcast.client.entities import Fact
from typing import Iterable
from pyfactcast.client.sync import FactStore


fact_store = FactStore()


def subscribe(
    namespace: str, follow: bool, from_now: bool = False, after_fact: UUID = None
) -> Iterable[Fact]:
    with fact_store as fs:
        return fs.subscribe(
            namespace=namespace,
            continuous=follow,
            from_now=from_now,
            after_fact=after_fact,
        )



from typing import Optional
from uuid import UUID
from pyfactcast.client.sync import FactStore


fact_store = FactStore()


def serial_of(fact_id: UUID) -> Optional[str]:
    with fact_store as fs:
        return fs.serial_of(fact_id=fact_id)



from dataclasses import dataclass
from typing import Optional

import logging
from os import environ as env


log_level = getattr(
    logging, env.get("FACTCAST_LOG_LEVEL", "WARNING").upper(), 30
)  # Setting a default here is pretty defensive. Anyhow nothing lost by doing it


@dataclass
class Credentials:
    username: str
    password: str

    def __post_init__(self):
        if self.username and not self.password:
            raise ValueError(
                "Username was provided without password. Please provide a non empty password."
            )
        if self.password and not self.username:
            raise ValueError(
                "Password was provided without a username. Please provide a non empty username."
            )
        if not self.username and not self.password:
            raise ValueError(
                "Both username and password are empty. Do not instantiate Credentials like this."
            )


@dataclass
class ClientConfiguration:
    server: str
    root_cert_path: Optional[str] = None
    ssl_target_override: Optional[str] = None
    credentials: Optional[Credentials] = None

    def __post_init__(self):
        if not self.server:
            raise ValueError("Server connection string missing.")


def get_client_configuration() -> ClientConfiguration:
    username = env.get("GRPC_USER", "")
    password = env.get("GRPC_PASSWORD", "")
    credentials = None

    if username or password:
        credentials = Credentials(username, password)

    return ClientConfiguration(
        server=env.get("GRPC_SERVER", ""),
        credentials=credentials,
        root_cert_path=env.get("GRPC_ROOT_CERT_PATH"),
        ssl_target_override=env.get("GRPC_CN_OVERWRITE"),
    )
